/*
 * Copyright (C) 2005-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file alsa_write.c paraslash's alsa output plugin */

/*
 * Based in parts on aplay.c from the alsa-utils-1.0.8 package,
 * Copyright (c) by Jaroslav Kysela <perex@suse.cz>, which is
 * based on the vplay program by Michael Beck.
 */

#include <regex.h>
#include <sys/types.h>
#include <alsa/asoundlib.h>
#include <sys/time.h>
#include <stdbool.h>

#include "para.h"
#include "fd.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "write.h"
#include "write_common.h"
#include "alsa_write.cmdline.h"
#include "error.h"

/** Data specific to the alsa writer. */
struct private_alsa_write_data {
	/** The alsa handle */
	snd_pcm_t *handle;
	/** Determined and set by alsa_init(). */
	int bytes_per_frame;
	/** The approximate maximum buffer duration in us. */
	unsigned buffer_time;
	/* Number of frames that fit into the buffer. */
	snd_pcm_uframes_t buffer_frames;
	/**
	 * The sample rate given by command line option or the decoder
	 * of the writer node group.
	 */
	unsigned sample_rate;
	snd_pcm_format_t sample_format;
	/**
	 * The number of channels, given by command line option or the
	 * decoder of the writer node group.
	 */
	unsigned channels;
	struct timeval drain_barrier;
};

static snd_pcm_format_t get_alsa_pcm_format(enum sample_format sf)
{
	switch (sf) {
	case SF_S8: return SND_PCM_FORMAT_S8;
	case SF_U8: return SND_PCM_FORMAT_U8;
	case SF_S16_LE: return SND_PCM_FORMAT_S16_LE;
	case SF_S16_BE: return SND_PCM_FORMAT_S16_BE;
	case SF_U16_LE: return SND_PCM_FORMAT_U16_LE;
	case SF_U16_BE: return SND_PCM_FORMAT_U16_BE;
	default: return SND_PCM_FORMAT_S16_LE;
	}
}

/* Install PCM software and hardware configuration. */
static int alsa_init(struct private_alsa_write_data *pad,
		struct alsa_write_args_info *conf)
{
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_sw_params_t *swparams;
	snd_pcm_uframes_t start_threshold, stop_threshold;
	snd_pcm_uframes_t period_size;
	int err;

	PARA_INFO_LOG("opening %s\n", conf->device_arg);
	err = snd_pcm_open(&pad->handle, conf->device_arg,
		SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK);
	if (err < 0)
		return -E_PCM_OPEN;

	snd_pcm_hw_params_alloca(&hwparams);
	snd_pcm_sw_params_alloca(&swparams);
	if (snd_pcm_hw_params_any(pad->handle, hwparams) < 0)
		return -E_BROKEN_CONF;
	if (snd_pcm_hw_params_set_access(pad->handle, hwparams,
			SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
		return -E_ACCESS_TYPE;
	if (snd_pcm_hw_params_set_format(pad->handle, hwparams,
			pad->sample_format) < 0)
		return -E_SAMPLE_FORMAT;
	if (snd_pcm_hw_params_set_channels(pad->handle, hwparams,
			pad->channels) < 0)
		return -E_CHANNEL_COUNT;
	if (snd_pcm_hw_params_set_rate_near(pad->handle, hwparams,
			&pad->sample_rate, NULL) < 0)
		return -E_SET_RATE;
	err = snd_pcm_hw_params_get_buffer_time_max(hwparams,
		&pad->buffer_time, NULL);
	if (err < 0 || !pad->buffer_time)
		return -E_GET_BUFFER_TIME;
	PARA_INFO_LOG("buffer time: %d\n", pad->buffer_time);
	if (snd_pcm_hw_params_set_buffer_time_near(pad->handle, hwparams,
			&pad->buffer_time, NULL) < 0)
		return -E_SET_BUFFER_TIME;
	if (snd_pcm_hw_params(pad->handle, hwparams) < 0)
		return -E_HW_PARAMS;
	snd_pcm_hw_params_get_period_size(hwparams, &period_size, NULL);
	snd_pcm_hw_params_get_buffer_size(hwparams, &pad->buffer_frames);
	PARA_INFO_LOG("buffer size: %lu, period_size: %lu\n", pad->buffer_frames,
		period_size);
	if (period_size == pad->buffer_frames)
		return -E_BAD_PERIOD;
	snd_pcm_sw_params_current(pad->handle, swparams);
	snd_pcm_sw_params_set_avail_min(pad->handle, swparams, period_size);
	if (pad->buffer_frames < 1)
		start_threshold = 1;
	else
		start_threshold = PARA_MIN(pad->buffer_frames,
			(snd_pcm_uframes_t)pad->sample_rate);
	if (snd_pcm_sw_params_set_start_threshold(pad->handle, swparams,
			start_threshold) < 0)
		return -E_START_THRESHOLD;
	stop_threshold = pad->buffer_frames;
	if (snd_pcm_sw_params_set_stop_threshold(pad->handle, swparams,
			stop_threshold) < 0)
		return -E_STOP_THRESHOLD;
	if (snd_pcm_sw_params(pad->handle, swparams) < 0)
		PARA_WARNING_LOG("unable to install sw params\n");
	pad->bytes_per_frame = snd_pcm_format_physical_width(pad->sample_format)
		* pad->channels / 8;
	if (pad->bytes_per_frame <= 0)
		return -E_PHYSICAL_WIDTH;
	PARA_INFO_LOG("bytes per frame: %d\n", pad->bytes_per_frame);
	if (snd_pcm_nonblock(pad->handle, 1))
		PARA_ERROR_LOG("failed to set nonblock mode\n");
	return 1;
}

static void alsa_write_pre_select(struct sched *s, struct task *t)
{
	struct writer_node *wn = container_of(t, struct writer_node, task);
	struct private_alsa_write_data *pad = wn->private_data;
	snd_pcm_sframes_t avail, underrun;
	int ret = btr_node_status(wn->btrn, wn->min_iqs, BTR_NT_LEAF);

	if (ret == 0)
		return;
	if (ret < 0 || !pad)
		return sched_min_delay(s);
	/*
	 * Data is available to be written to the alsa handle.  Compute number
	 * of milliseconds until next buffer underrun would occur.
	 *
	 * snd_pcm_avail_update() updates the current available count of
	 * samples for writing.  It is a light method to obtain current stream
	 * position, because it does not require the user <-> kernel context
	 * switch, but the value is less accurate, because ring buffer pointers
	 * are updated in kernel drivers only when an interrupt occurs.
	 */
	avail = snd_pcm_avail_update(pad->handle);
	if (avail < 0)
		avail = 0;
	underrun = (pad->buffer_frames - avail) * pad->buffer_time
		/ pad->buffer_frames / 1000;
	if (underrun < 50)
		underrun = 50;
	underrun -= 50;
	sched_request_timeout_ms(underrun, s);
}

static void alsa_close(struct writer_node *wn)
{
	struct private_alsa_write_data *pad = wn->private_data;
	PARA_INFO_LOG("closing writer node %p\n", wn);

	if (!pad)
		return;
	/*
	 * It's OK to have a blocking operation here because we already made
	 * sure that the PCM output buffer is (nearly) empty.
	 */
	snd_pcm_nonblock(pad->handle, 0);
	snd_pcm_drain(pad->handle);
	snd_pcm_close(pad->handle);
	snd_config_update_free_global();
	free(pad);
}

static void alsa_write_post_select(__a_unused struct sched *s,
		struct task *t)
{
	struct writer_node *wn = container_of(t, struct writer_node, task);
	struct private_alsa_write_data *pad = wn->private_data;
	struct btr_node *btrn = wn->btrn;
	char *data;
	size_t bytes;
	snd_pcm_sframes_t frames;
	int ret;

again:
	t->error = 0;
	ret = btr_node_status(btrn, wn->min_iqs, BTR_NT_LEAF);
	if (ret == 0)
		return;
	btr_merge(btrn, wn->min_iqs);
	bytes = btr_next_buffer(btrn, &data);
	if (ret < 0 || bytes < wn->min_iqs) { /* eof */
		assert(btr_no_parent(btrn));
		ret = -E_WRITE_COMMON_EOF;
		if (!pad)
			goto err;
		/* wait until pending frames are played */
		if (pad->drain_barrier.tv_sec == 0) {
			PARA_DEBUG_LOG("waiting for device to drain\n");
			tv_add(now, &(struct timeval)EMBRACE(0, 200 * 1000),
				&pad->drain_barrier);
			return;
		}
		if (tv_diff(now, &pad->drain_barrier, NULL) > 0)
			goto err;
		return;
	}
	if (!pad) {
		int32_t val;

		if (bytes == 0) /* no data available */
			return;
		pad = para_calloc(sizeof(*pad));
		get_btr_sample_rate(btrn, &val);
		pad->sample_rate = val;
		get_btr_channels(btrn, &val);
		pad->channels = val;
		get_btr_sample_format(btrn, &val);
		pad->sample_format = get_alsa_pcm_format(val);

		PARA_INFO_LOG("%d channel(s), %dHz\n", pad->channels,
			pad->sample_rate);
		ret = alsa_init(pad, wn->conf);
		if (ret < 0) {
			free(pad);
			goto err;
		}
		wn->private_data = pad;
		wn->min_iqs = pad->bytes_per_frame;
		goto again;
	}
	frames = bytes / pad->bytes_per_frame;
	frames = snd_pcm_writei(pad->handle, data, frames);
	if (frames >= 0) {
		btr_consume(btrn, frames * pad->bytes_per_frame);
		goto again;
	}
	if (frames == -EPIPE) {
		PARA_WARNING_LOG("underrun (tried to write %zu bytes)\n", bytes);
		snd_pcm_prepare(pad->handle);
		return;
	}
	if (frames == -EAGAIN)
		return;
	PARA_WARNING_LOG("%s\n", snd_strerror(-frames));
	ret = -E_ALSA_WRITE;
err:
	assert(ret < 0);
	btr_remove_node(btrn);
	t->error = ret;
}

__malloc static void *alsa_parse_config_or_die(const char *options)
{
	struct alsa_write_args_info *conf = para_calloc(sizeof(*conf));

	PARA_INFO_LOG("options: %s, %zd\n", options, strcspn(options, " \t"));
	/* exits on errors */
	alsa_cmdline_parser_string(options, conf, "alsa_write");
	return conf;
}

static void alsa_free_config(void *conf)
{
	alsa_cmdline_parser_free(conf);
}

/**
 * The init function of the alsa writer.
 *
 * \param w Pointer to the writer to initialize.
 *
 * \sa struct \ref writer.
 */
void alsa_write_init(struct writer *w)
{
	struct alsa_write_args_info dummy;

	alsa_cmdline_parser_init(&dummy);
	w->close = alsa_close;
	w->pre_select = alsa_write_pre_select;
	w->post_select = alsa_write_post_select;
	w->parse_config_or_die = alsa_parse_config_or_die;
	w->shutdown = NULL; /* nothing to do */
	w->free_config = alsa_free_config;
	w->help = (struct ggo_help) {
		.short_help = alsa_write_args_info_help,
		.detailed_help = alsa_write_args_info_detailed_help
	};
	alsa_cmdline_parser_free(&dummy);
}
