/*
 * Copyright (C) 2005-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
/** \file udp_recv.c Paraslash's udp receiver */

#include <regex.h>
#include <sys/socket.h>
#include <net/if.h>

#include "para.h"
#include "error.h"
#include "portable_io.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "recv.h"
#include "udp_recv.cmdline.h"
#include "string.h"
#include "net.h"
#include "fd.h"
#include "buffer_tree.h"

/**
 * Data specific to the udp receiver.
 *
 * \sa \ref receiver, \ref receiver_node.
 */
struct private_udp_recv_data {
	/** The socket file descriptor. */
	int fd;
	struct btr_pool *btrp;
};

static void udp_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_udp_recv_data *purd = rn->private_data;

	if (generic_recv_pre_select(s, t) <= 0)
		return;
	para_fd_set(purd->fd, &s->rfds, &s->max_fileno);
}

static int udp_check_eof(size_t sz, struct iovec iov[2])
{
	if (sz < FEC_EOF_PACKET_LEN)
		return 0;
	if (iov[0].iov_len >= FEC_EOF_PACKET_LEN) {
		if (memcmp(iov[0].iov_base, FEC_EOF_PACKET,
				FEC_EOF_PACKET_LEN) != 0)
			return 0;
		return -E_RECV_EOF;
	}
	if (memcmp(iov[0].iov_base, FEC_EOF_PACKET, iov[0].iov_len) != 0)
		return 0;
	if (memcmp(iov[1].iov_base, FEC_EOF_PACKET + iov[0].iov_len,
			FEC_EOF_PACKET_LEN - iov[0].iov_len) != 0)
		return 0;
	return -E_RECV_EOF;
}

static void udp_recv_post_select(__a_unused struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_udp_recv_data *purd = rn->private_data;
	struct btr_node *btrn = rn->btrn;
	size_t num_bytes;
	struct iovec iov[2];
	int ret, readv_ret, iovcnt;

	t->error = 0;
	ret = btr_node_status(btrn, 0, BTR_NT_ROOT);
	if (ret <= 0)
		goto out;
	iovcnt = btr_pool_get_buffers(purd->btrp, iov);
	ret = -E_UDP_OVERRUN;
	if (iovcnt == 0)
		goto out;
	ret = readv_nonblock(purd->fd, iov, iovcnt, &s->rfds, &num_bytes);
	if (num_bytes == 0)
		goto out;
	readv_ret = ret;
	ret = udp_check_eof(num_bytes, iov);
	if (ret < 0)
		goto out;
	if (iov[0].iov_len >= num_bytes)
		btr_add_output_pool(purd->btrp, num_bytes, btrn);
	else { /* both buffers contain data */
		btr_add_output_pool(purd->btrp, iov[0].iov_len, btrn);
		btr_add_output_pool(purd->btrp, num_bytes - iov[0].iov_len,
			btrn);
	}
	ret = readv_ret;
out:
	if (ret >= 0)
		return;
	btr_remove_node(btrn);
	t->error = ret;
	close(purd->fd);
	purd->fd = -1;
}

static void udp_recv_close(struct receiver_node *rn)
{
	struct private_udp_recv_data *purd = rn->private_data;

	if (purd->fd >= 0)
		close(purd->fd);
	btr_pool_free(purd->btrp);
	free(rn->private_data);
}

static void *udp_recv_parse_config(int argc, char **argv)
{
	int ret;
	struct udp_recv_args_info *tmp =
		para_calloc(sizeof(struct udp_recv_args_info));

	ret = udp_recv_cmdline_parser(argc, argv, tmp)? -E_UDP_SYNTAX : 1;
	if (ret >= 0)
		return tmp;
	free(tmp);
	return NULL;
}

/*
 * Perform AF-independent joining of multicast receive addresses.
 *
 * \param fd	Bound socket descriptor.
 * \param iface	The receiving multicast interface, or NULL for the default.
 *
 * \return Zero if okay, negative on error.
 */
static int mcast_receiver_setup(int fd, const char *iface)
{
	struct sockaddr_storage ss;
	socklen_t sslen = sizeof(ss);
	int id = iface == NULL ? 0 : if_nametoindex(iface);

	if (getsockname(fd, (struct sockaddr *)&ss, &sslen) < 0)
		goto err;
	assert(ss.ss_family == AF_INET || ss.ss_family == AF_INET6);

	if (iface != NULL && id == 0)
		PARA_WARNING_LOG("could not resolve interface %s, using default\n",
			iface);

	switch (ss.ss_family) {
	case AF_INET:
		if (IN_MULTICAST(htonl(((struct sockaddr_in *)&ss)->sin_addr.s_addr))) {
#ifdef HAVE_IP_MREQN
			struct ip_mreqn m4;

			m4.imr_address.s_addr	= INADDR_ANY;
			m4.imr_ifindex		= id;
#else
			struct ip_mreq m4;

			m4.imr_interface.s_addr	= INADDR_ANY;
			if (id != 0)
				PARA_ERROR_LOG("Setting IPv4 receiver mcast interface not supported\n");

#endif
			m4.imr_multiaddr = ((struct sockaddr_in *)&ss)->sin_addr;

			if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
					&m4, sizeof(m4)) < 0)
				break;
		}
		return 0;
	case AF_INET6:
		if (IN6_IS_ADDR_MULTICAST(&((struct sockaddr_in6 *)&ss)->sin6_addr)) {
			struct ipv6_mreq m6;

			memset(&m6, 0, sizeof(m6));
			memcpy(&m6.ipv6mr_multiaddr, &((struct sockaddr_in6 *)&ss)->sin6_addr, 16);
			m6.ipv6mr_interface = id;
			if (setsockopt(fd, IPPROTO_IPV6, IPV6_JOIN_GROUP, &m6, sizeof(m6)) < 0)
				break;
		}
		return 0;
	}
err:
	return -ERRNO_TO_PARA_ERROR(errno);
}

static int udp_recv_open(struct receiver_node *rn)
{
	struct private_udp_recv_data *purd;
	struct udp_recv_args_info *c = rn->conf;
	char  *iface = c->iface_given ? c->iface_arg : NULL;
	int ret;

	rn->private_data = para_calloc(sizeof(struct private_udp_recv_data));
	purd = rn->private_data;

	ret = makesock(IPPROTO_UDP, 1, c->host_arg, c->port_arg, NULL);
	if (ret < 0)
		goto err;
	purd->fd = ret;

	ret = mcast_receiver_setup(purd->fd, iface);
	if (ret < 0) {
		close(purd->fd);
		goto err;
	}

	ret = mark_fd_nonblocking(purd->fd);
	if (ret < 0) {
		close(purd->fd);
		goto err;
	}
	PARA_INFO_LOG("receiving from %s:%d, fd=%d\n", c->host_arg,
		c->port_arg, purd->fd);
	purd->btrp = btr_pool_new("udp_recv", 320 * 1024);
	return purd->fd;
err:
	free(rn->private_data);
	return ret;
}

static void udp_recv_free_config(void *conf)
{
	udp_recv_cmdline_parser_free(conf);
	free(conf);
}

/**
 * The init function of the udp receiver.
 *
 * \param r Pointer to the receiver struct to initialize.
 *
 * Initialize all function pointers of \a r.
 */
void udp_recv_init(struct receiver *r)
{
	struct udp_recv_args_info dummy;

	udp_recv_cmdline_parser_init(&dummy);
	r->open = udp_recv_open;
	r->close = udp_recv_close;
	r->pre_select = udp_recv_pre_select;
	r->post_select = udp_recv_post_select;
	r->parse_config = udp_recv_parse_config;
	r->free_config = udp_recv_free_config;
	r->help = (struct ggo_help) {
		.short_help = udp_recv_args_info_help,
		.detailed_help = udp_recv_args_info_detailed_help
	};
	udp_recv_cmdline_parser_free(&dummy);
}
