/*
 * Copyright (C) 2006-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write_common.c common functions of para_audiod and para_write */

#include <regex.h>
#include <stdbool.h>

#include "para.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "write.h"
#include "error.h"

/** the array containing the names of all supported writers */
const char *writer_names[] ={WRITER_NAMES};

/** the array of supported writers */
struct writer writers[NUM_SUPPORTED_WRITERS] = {WRITER_ARRAY};

/**
 * Call the init function of each supported paraslash writer.
 */
void writer_init(void)
{
	int i;

	FOR_EACH_WRITER(i)
		writers[i].init(&writers[i]);
}
/**
 * Check if given string is a valid command line for any writer.
 *
 * \param \wa String of the form writer_name:options.
 * \param writer_num Contains the number of the writer upon success.
 *
 * This function checks whether \a wa starts with the name of a supported
 * paraslash writer, optionally followed by a colon and any options for that
 * writer.  If a valid writer name was found and further are present, the
 * remaining part of \a wa is passed to that writer's config parser.
 *
 * \return On success, a pointer to the gengetopt args info struct is returned
 * and \a writer_num contains the number of the writer. Otherwise this function
 * returns \p NULL.
 */
void *check_writer_arg(const char *wa, int *writer_num)
{
	int i;

	*writer_num = -E_WRITE_COMMON_SYNTAX;
	PARA_INFO_LOG("checking  %s\n", wa);
	FOR_EACH_WRITER(i) {
		const char *name = writer_names[i];
		size_t len = strlen(name);
		char c;
		if (strlen(wa) < len)
			continue;
		if (strncmp(name, wa, len))
			continue;
		c = wa[len];
		if (c && c != ' ')
			continue;
		*writer_num = i;
		return writers[i].parse_config_or_die(c? wa + len + 1 : "");
	}
	PARA_ERROR_LOG("writer not found\n");
	return NULL;
}

/**
 * Open a writer node and register the corresponding task.
 *
 * \param wn The writer node to open.
 * \param parent The parent btr node (the source for the writer node).
 *
 * The configuration of the writer node stored in \p wn->conf must be
 * initialized before this function may be called.
 */
void register_writer_node(struct writer_node *wn, struct btr_node *parent)
{
	struct writer *w = writers + wn->writer_num;
	char *name = make_message("%s writer", writer_names[wn->writer_num]);

	wn->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = name, .parent = parent,
		.handler = w->execute, .context = wn));
	strcpy(wn->task.status, name);
	free(name);
	wn->task.post_select = w->post_select;
	wn->task.pre_select = w->pre_select;
	register_task(&wn->task);
}

/**
 * Parse config and register a task for a writer node.
 *
 * \param arg Command line arguments.
 * \param parent The new node will be a child of \a parent.
 * \param wn The writer node.
 *
 * If arg is \p NULL, the OS-dependent default writer is used with no
 * arguments.  The default writers are alsa for Linux, osx for OS X, oss for
 * *BSD, and the file writer if the default writer is not supported.
 *
 * Once the writer configuration has been retrieved from the ->parse_config
 * callback a writer node is created, its buffer tree node is added to the
 * buffer tree as a child of the given parent.
 *
 * Finally, the new writer node's task structure is initialized and registered
 * to the paraslash scheduler.
 *
 * \return Standard.
 */
int setup_writer_node(const char *arg, struct btr_node *parent,
		struct writer_node *wn)
{
	if (arg)
		wn->conf = check_writer_arg(arg, &wn->writer_num);
	else {
		wn->writer_num = DEFAULT_WRITER;
		wn->conf = writers[DEFAULT_WRITER].parse_config_or_die("");
	}
	if (!wn->conf)
		return -E_WRITE_COMMON_SYNTAX;
	register_writer_node(wn, parent);
	return 1;
}

/**
 * Print the help text of all writers to stdout.
 *
 * \param detailed Whether to print the detailed help text.
 */
void print_writer_helps(int detailed)
{
	int i;

	printf_or_die("\nAvailable writers: \n\t");
	FOR_EACH_WRITER(i)
		printf_or_die("%s%s", i? " " : "", writer_names[i]);
	printf_or_die("\n\n");
	FOR_EACH_WRITER(i) {
		struct writer *w = writers + i;

		if (!w->help.short_help)
			continue;
		printf_or_die("Options for %s:\n", writer_names[i]);
		ggo_print_help(&w->help, detailed);
	}
}

static void get_btr_value(struct btr_node *btrn, const char *cmd,
		int32_t *result)
{
	char *buf = NULL;
	int ret = btr_exec_up(btrn, cmd, &buf);

	if (ret < 0) {
		/*
		 * This really should not happen. It means one of our parent
		 * nodes died unexpectedly. Proceed with fingers crossed.
		 */
		PARA_CRIT_LOG("cmd %s: %s\n", cmd, para_strerror(-ret));
		*result = 0;
		return;
	}
	ret = para_atoi32(buf, result);
	assert(ret >= 0);
	free(buf);
}

/**
 * Ask parent btr nodes for the sample rate of the current stream.
 *
 * \param btrn Where to start the search.
 * \param result Filled in by this function.
 *
 * This function is assumed to succeed and terminates on errors.
 */
void get_btr_sample_rate(struct btr_node *btrn, int32_t *result)
{
	get_btr_value(btrn, "sample_rate", result);
}

/**
 * Ask parent btr nodes for the channel count of the current stream.
 *
 * \param btrn See \ref get_btr_sample_rate.
 * \param result See \ref get_btr_sample_rate.
 */
void get_btr_channels(struct btr_node *btrn, int32_t *result)
{
	get_btr_value(btrn, "channels", result);
}

/**
 * Ask parent btr nodes for the number of bits per sample and the byte sex.
 *
 * \param btrn See \ref get_btr_sample_rate.
 * \param result Contains the sample format as an enum sample_format type.
 */
void get_btr_sample_format(struct btr_node *btrn, int32_t *result)
{
	get_btr_value(btrn, "sample_format", result);
}
